/* mapping table of MathML elements */
#include "MathML.h"
#include "MathMLentities.h"
static ElemMapping    MathMLElemMappingTable[] =
{
   /* This table MUST be in alphabetical order */
   {"XMLPI", SPACE, MathML_EL_XMLPI, L_Other, FALSE},
   {"XMLPI_line", SPACE, MathML_EL_XMLPI_line, L_Other, FALSE},
   {"XMLcomment", SPACE, MathML_EL_XMLcomment, L_Other, FALSE},
   {"XMLcomment_line", SPACE, MathML_EL_XMLcomment_line, L_Other, FALSE},
   {"c", SPACE, MathML_EL_TEXT_UNIT, L_Other, TRUE},
   {"maligngroup", 'E', MathML_EL_MALIGNGROUP, L_Other, FALSE},
   {"malignmark", 'E', MathML_EL_MALIGNMARK, L_Other, FALSE},
   {"math", SPACE, MathML_EL_MathML, L_Other, FALSE},
   {"menclose", SPACE, MathML_EL_MENCLOSE, L_Other, FALSE},
   {"merror", SPACE, MathML_EL_MERROR, L_Other, FALSE},
   {"mf", SPACE, MathML_EL_MF, L_Other, FALSE},  /* for compatibility with an old version of MathML: WD-math-970704 */
   {"mfenced", SPACE, MathML_EL_MFENCED, L_Other, FALSE},
   {"mfrac", SPACE, MathML_EL_MFRAC, L_Other, FALSE},
   {"mglyph", 'E', MathML_EL_MGLYPH, L_Other, FALSE},
   {"mi", SPACE, MathML_EL_MI, L_Other, FALSE},
   {"mlabeledtr", SPACE, MathML_EL_MLABELEDTR, L_Other, FALSE},
   {"mmultiscripts", SPACE, MathML_EL_MMULTISCRIPTS, L_Other, FALSE},
   {"mn", SPACE, MathML_EL_MN, L_Other, FALSE},
   {"mo", SPACE, MathML_EL_MO, L_Other, FALSE},
   {"mover", SPACE, MathML_EL_MOVER, L_Other, FALSE},
   {"mpadded", SPACE, MathML_EL_MPADDED, L_Other, FALSE},
   {"mphantom", SPACE, MathML_EL_MPHANTOM, L_Other, FALSE},
   {"mprescripts", SPACE, MathML_EL_PrescriptPairs, L_Other, FALSE},
   {"mroot", SPACE, MathML_EL_MROOT, L_Other, FALSE},
   {"mrow", SPACE, MathML_EL_MROW, L_Other, FALSE},
   {"ms", SPACE, MathML_EL_MS, L_Other, FALSE},
   {"mspace", 'E', MathML_EL_MSPACE, L_Other, FALSE},
   {"msqrt", SPACE, MathML_EL_MSQRT, L_Other, FALSE},
   {"mstyle", SPACE, MathML_EL_MSTYLE, L_Other, FALSE},
   {"msub", SPACE, MathML_EL_MSUB, L_Other, FALSE},
   {"msubsup", SPACE, MathML_EL_MSUBSUP, L_Other, FALSE},
   {"msup", SPACE, MathML_EL_MSUP, L_Other, FALSE},
   {"mtable", SPACE, MathML_EL_MTABLE, L_Other, FALSE},
   {"mtd", SPACE, MathML_EL_MTD, L_Other, FALSE},
   {"mtext", SPACE, MathML_EL_MTEXT, L_Other, FALSE},
   {"mtr", SPACE, MathML_EL_MTR, L_Other, FALSE},
   {"munder", SPACE, MathML_EL_MUNDER, L_Other, FALSE},
   {"munderover", SPACE, MathML_EL_MUNDEROVER, L_Other, FALSE},
   {"none", SPACE, MathML_EL_Construct, L_Other, FALSE},
   {"", SPACE, 0, L_Other, FALSE}	 /* Last entry. Mandatory */
};

static AttributeMapping MathMLAttributeMappingTable[] =
{
   /* The first entry MUST be unknown_attr */
   /* The rest of this table MUST be in alphabetical order */
   {"unknown_attr", "", 'A', MathML_ATTR_Invalid_attribute, L_Other},

   {"accent", "", 'A', MathML_ATTR_accent, L_Other},
   {"accentunder", "", 'A', MathML_ATTR_accentunder, L_Other},
   {"actiontype", "", 'A', MathML_ATTR_actiontype, L_Other},
   {"align", "", 'A', MathML_ATTR_align, L_Other},
   {"alignmentscope", "", 'A', MathML_ATTR_alignmentscope, L_Other},
   {"alt", "", 'A', MathML_ATTR_alt, L_Other},
   {"altimg", "", 'A', MathML_ATTR_altimg, L_Other},
   {"alttext", "", 'A', MathML_ATTR_alttext, L_Other},
   {"background", "", 'A', MathML_ATTR_background_, L_Other},
   {"bevelled", "", 'A', MathML_ATTR_bevelled, L_Other},
   {"class", "", 'A', MathML_ATTR_class, L_Other},
   {"close", "", 'A', MathML_ATTR_close, L_Other},
   {"columnalign", "mtd", 'A', MathML_ATTR_columnalign_mtd, L_Other},
   {"columnalign", "", 'A', MathML_ATTR_columnalign, L_Other},
   {"columnlines", "", 'A', MathML_ATTR_columnlines, L_Other},
   {"columnspacing", "", 'A', MathML_ATTR_columnspacing, L_Other},
   {"columnspan", "", 'A', MathML_ATTR_columnspan, L_Other},
   {"columnwidth", "", 'A', MathML_ATTR_columnwidth, L_Other},
   {"color", "", 'A', MathML_ATTR_color, L_Other},
   {"denomalign", "", 'A', MathML_ATTR_denomalign, L_Other},
   {"depth", "", 'A', MathML_ATTR_depth_, L_Other},
   {"display", "", 'A', MathML_ATTR_display, L_Other},
   {"displaystyle", "", 'A', MathML_ATTR_displaystyle, L_Other},
   {"edge", "", 'A', MathML_ATTR_edge, L_Other},
   {"equalcolumns", "", 'A', MathML_ATTR_equalcolumns, L_Other},
   {"equalrows", "", 'A', MathML_ATTR_equalrows, L_Other},
   {"fence", "", 'A', MathML_ATTR_fence, L_Other},
   {"fontfamily", "", 'A', MathML_ATTR_fontfamily, L_Other},
   {"fontstyle", "", 'A', MathML_ATTR_fontstyle, L_Other},
   {"fontsize", "", 'A', MathML_ATTR_fontsize, L_Other},
   {"fontweight", "", 'A', MathML_ATTR_fontweight, L_Other},
   {"form", "", 'A', MathML_ATTR_form, L_Other},
   {"frame", "", 'A', MathML_ATTR_frame, L_Other},
   {"framespacing", "", 'A', MathML_ATTR_framespacing, L_Other},
   {"groupalign", "maligngroup", 'A', MathML_ATTR_groupalign_malgr, L_Other},
   {"groupalign", "", 'A', MathML_ATTR_groupalign, L_Other},
   {"height", "", 'A', MathML_ATTR_height_, L_Other},
   {"id", "", 'A', MathML_ATTR_id, L_Other},
   {"index", "", 'A', MathML_ATTR_index, L_Other},
   {"largeop", "", 'A', MathML_ATTR_largeop, L_Other},
   {"linebreak", "", 'A', MathML_ATTR_linebreak_, L_Other},
   {"linethickness", "", 'A', MathML_ATTR_linethickness, L_Other},
   {"lquote", "", 'A', MathML_ATTR_lquote, L_Other},
   {"lspace", "", 'A', MathML_ATTR_lspace, L_Other},
   {"mathbackground", "", 'A', MathML_ATTR_mathbackground, L_Other},
   {"mathcolor", "", 'A', MathML_ATTR_mathcolor, L_Other},
   {"mathsize", "", 'A', MathML_ATTR_mathsize, L_Other},
   {"mathvariant", "", 'A', MathML_ATTR_mathvariant, L_Other},
   {"maxsize", "", 'A', MathML_ATTR_maxsize, L_Other},
   {"minlabelspacing", "", 'A', MathML_ATTR_minlabelspacing, L_Other},
   {"minsize", "", 'A', MathML_ATTR_minsize, L_Other},
   {"mode", "", 'A', MathML_ATTR_display, L_Other},
   {"movablelimits", "", 'A', MathML_ATTR_movablelimits, L_Other},
   {"notation", "", 'A', MathML_ATTR_notation, L_Other},
   {"numalign", "", 'A', MathML_ATTR_numalign, L_Other},
   {"open", "", 'A', MathML_ATTR_open, L_Other},
   {"other", "", 'A', MathML_ATTR_other, L_Other},
   {"overflow", "", 'A', MathML_ATTR_overflow, L_Other},
   {"rowalign", "mtr", 'A', MathML_ATTR_rowalign_mtr, L_Other},
   {"rowalign", "mlabeledtr", 'A', MathML_ATTR_rowalign_mtr, L_Other},
   {"rowalign", "mtd", 'A', MathML_ATTR_rowalign_mtr, L_Other},
   {"rowalign", "", 'A', MathML_ATTR_rowalign, L_Other},
   {"rowlines", "", 'A', MathML_ATTR_rowlines, L_Other},
   {"rowspacing", "", 'A', MathML_ATTR_rowspacing, L_Other},
   {"rowspan", "", 'A', MathML_ATTR_rowspan_, L_Other},
   {"rquote", "", 'A', MathML_ATTR_rquote, L_Other},
   {"rspace", "", 'A', MathML_ATTR_rspace, L_Other},
   {"scriptlevel", "", 'A', MathML_ATTR_scriptlevel, L_Other},
   {"scriptminsize", "", 'A', MathML_ATTR_scriptminsize, L_Other},
   {"scriptsizemultiplier", "", 'A', MathML_ATTR_scriptsizemultiplier, L_Other},
   {"selection", "", 'A', MathML_ATTR_selection, L_Other},
   {"separator", "", 'A', MathML_ATTR_separator, L_Other},
   {"separators", "", 'A', MathML_ATTR_separators, L_Other},
   {"side", "", 'A', MathML_ATTR_side, L_Other},
   {"stretchy", "", 'A', MathML_ATTR_stretchy, L_Other},
   {"style", "", 'A', MathML_ATTR_style_, L_Other},
   {"subscriptshift", "", 'A', MathML_ATTR_subscriptshift, L_Other},
   {"superscriptshift", "", 'A', MathML_ATTR_superscriptshift, L_Other},
   {"symmetric", "", 'A', MathML_ATTR_symmetric, L_Other},
   {"width", "", 'A', MathML_ATTR_width_, L_Other},
   {"xml:space", "", 'A', MathML_ATTR_xml_space, L_Other},
   {"xref", "", 'A', MathML_ATTR_xref, L_Other},
   {"zzghost", "", 'A', MathML_ATTR_Ghost_restruct, L_Other},

   {"", "", EOS, 0, L_Other}		/* Last entry. Mandatory */
};

